/***************************************************************************
 *                                                                         *
 *                  (begin: Feb 20 2003)                                   *
 *                                                                         *
 *   Parallel IQPNNI - Important Quartet Puzzle with NNI                   *
 *                                                                         *
 *   Copyright (C) 2005 by Le Sy Vinh, Bui Quang Minh, Arndt von Haeseler  *
 *   Copyright (C) 2003-2004 by Le Sy Vinh, Arndt von Haeseler             *
 *   {vinh,minh}@cs.uni-duesseldorf.de                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef INTERFACE_H
#define INTERFACE_H

#include <iostream>
#include <fstream>

#include "model.h"
#include "ali.h"
#include "rate.h"
#include "outstream.h"
#include "ptnls.h"
#include "model.h"
#include "urtree.h"

#ifdef PARALLEL
#	include <mpi.h>
#endif // PARALLEL

const int PROG_ALL            = 0;
const int PROG_COMPUTE_MATRIX = 1;
const int PROG_PARM_ESTIMATE  = 2;

extern int program_mode;
extern bool load_check_point;
extern bool major_step_test;
extern bool nni_test;


const MODEL DEFAULT_SUBST_MODEL = HKY85;

struct CommonParameters {
    int        nIter;
    double     probDel;
    int        nRep;
    STOPPING_TYPE     stoppingRule;
    DATA_TYPE  dataType;
    MODEL      model;
    double      baseA, baseC, baseG, baseT;
    PAM_TYPE    baseFrqType;

    double      tsTvRatio;
    PAM_TYPE    tsTvRatioType;

    double      pyPuRatio;
    PAM_TYPE    pyPuRatioType;

    double      tsAG;
    double      tsCT;
    double      tvAC;
    double      tvAT;
    double      tvCG;
    double      tvGT;
    PAM_TYPE    genPamType;
    RATE_TYPE   rateType;
    int         nRate;
    double      gammaShape;
    PAM_TYPE    gammaShapeType;
    double 		prob_invar;
    PAM_TYPE   	prob_invar_type;
    bool 		invariable_site;
	
	// codon
	MODEL 		codon_model;
    PAM_TYPE    codonFrqType;
    double      codon_tsTvRatio;
    PAM_TYPE    codon_tsTvRatioType;
    NSSITE_TYPE nsSy_ratio_type;
    int        	nsSy_classes;
	double 		nsSy_ratio_val[MAX_NUM_CLASS];
	double 		nsSy_prob_val[MAX_NUM_CLASS];

    int         outGrpSeqNo;
	int 		pam_brent; // 2 for BRENT method to optimize parameters
							// 1 for Simulated Annealing + BFGS
							// 0 for BFGS algorithm only
};

void initializeParams(CommonParameters &params);


class Interface {

public :

	int cmpnIter ();

	double cmpProbSel (PAM_TYPE sourceType);

	void getIntNumber (const char *msg, double lowerBound, double upperBound, int &value);

	void getDoubleNumber (const char *msg, double lowerBound, double upperBound, double &value);

	//=====================================================================================
	//=====================================================================================
	int getOutGrpSeq ();

	/*=========================================================================================
	  Get the parameter by interactive interface
	//=========================================================================================*/
	int getParameter (bool interactive, CommonParameters &params);

	int getCommandLine( int argc, char *argv[],
                            char *userTreeFile,
                            char   *aliFileName,
                            CommonParameters &params,
							bool &interactive_interface);

	//====================================================================================
	//ab:cccc:dd: X
	int getIntNum (const char  *aLine);

	//====================================================================================
	//ab:cccc:dd: X
	double getDoubleNum (const char  *aLine);

	//====================================================================================
	//====================================================================================
	//get all parameters from the last stopped point
	int getPamCheckPoint ( char          *checkPointFileName, CommonParameters &params);

	//====================================================================================
	//====================================================================================
	int isExistedFile (const char *fileName);

	/*====================================================================================
	 
	Interactive interface to get the user paramemters
	 
	====================================================================================*/
	int getUserPam ( int argc, char *argv[],
	                 char *userTreeFile,
	                 char   *aliFileName,
					 CommonParameters &params);

	/*****************************************************************
	Print the usage dialog  
	****************************************************************/
	void usage();

	//=================================================================
	//=================================================================
	void writeSpecificRate (const char *rateFileName);

	//=====================================================================================================
	//=====================================================================================================
	//=====================================================================================================
	void writeOut (const char *aliFile,
	               UrTree &tree,
	               int nIter, double probSel, int nRep,
	               STOPPING_TYPE stoppingRule,
	               Vec<char> &conTree,
	               const char *startedDate, double progTime);


private:
	int nSeq_;
};
#endif

