/***************************************************************************
 *                                                                         *
 *                  (begin: Feb 20 2003)                                   *
 *                                                                         *
 *   Parallel IQPNNI - Important Quartet Puzzle with NNI                   *
 *                                                                         *
 *   Copyright (C) 2005 by Le Sy Vinh, Bui Quang Minh, Arndt von Haeseler  *
 *   Copyright (C) 2003-2004 by Le Sy Vinh, Arndt von Haeseler             *
 *   {vinh,minh}@cs.uni-duesseldorf.de                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
                          model.h  -  description
                             -------------------
    begin                : Wed Mar 5 2003
    copyright            : (C) 2003 by 
    email                : vinh@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef MODEL_H
#define MODEL_H

#include "constant.h"
#include "mat.h"
#include "basem.h"

#include "gtrm.h"

#include "hky85m.h"
#include "tn93m.h"
#include "codonmodel.h"
#include "codonny98.h"
#include "codonyn98.h"
#include "codonpedersen98.h"
#include "codongtr.h"
#include "codonposrate.h"

#include "aminom.h"
#include "dvec20.h"
#include "dmat20.h"

//this class is an interface for all methods
class Model {
public:
	//the constructor function
	Model();


	int getNState ();


	//set the type of evolutional change model
	void setModelType (const MODEL modelType);
	void setDataType (DATA_TYPE dataType);
	void setTsTvRatioType (PAM_TYPE tsTvRatioType);
	void setPyPuRatioType (PAM_TYPE pyPuRatioType);

	void setPyPuRatio (PAM_TYPE pyPuRatio);
	void setGenPamType (PAM_TYPE genPamType);
	void setBaseFrqType (PAM_TYPE baseFrqType);
	
	int isEstedPam ();


	//get the type of evolutional change model
	MODEL getModelType ();

	//get the type of evolutional change model
	DATA_TYPE getDataType ();

	bool isCodonAnalysisNext();
	
	PAM_TYPE getTsTvRatioType ();
	PAM_TYPE getPyPuRatioType ();
	PAM_TYPE getGenPamType ();
	PAM_TYPE getBaseFrqType ();

	//set the transition transvertion ratio
	void setTsTvRatio (const double tsTvRatio);
	void setPyPuRatio (const double pyPuRatio);
	//void setVParm(const double v_parm);

	void setGenPam (const double tsAG, const double tsCT,
	                const double tvAC, const double tvAT, const double tvCG, const double tvGT);


	double getTsTvRatio ();
	double getPyPuRatio ();


	//everything is inited here
	void init ();
	void setNState (int nState);

	void setStateFrq (double stateFrqA, double stateFrqC, double stateFrqG, double stateFrqT);

	//return the state frequence
	void getStateFrq (DVec20 &stateFrqArr);
	//================================================================
	void getGenPam (double &tsAG, double & tsCT, double & tvAC, double & tvAT, double & tvCG, double & tvGT);


	//return the state frequence
	double getStateFrq (const int stateNo);
	MODEL getType ();

	/**
		get the propotion of a class
	*/
	double getClassProb(int classNo);
	
	/**
		get the ratio parm of a class
	*/
	double getClassRatio(int classNo);

	/*optimize all parameter evolving to sequence evolution
	1. tsTvRatio
	 
	2. TsRatio
	3. find state frequencies
	*/
	int optPam ();

	/******************************************************
	* compute the probability changes 
	*******************************************************/

	//compute the probability mat
	void cmpProbChange (const double time, DMat20 &probMat);

	//compute the probability mat, for codon-based Nielsen Yang 98 Model
	void cmpProbChange (const double time, const int classNo, DMat20 &probMat);
	
	//compute the first order probability mat

	//compute the probability vec
	void cmpProbChange (const int stateNo, const double time, DVec20 &probVec);
	//compute the first order probability vec


	//compute the probability of changing from stateNo1 into stateNo2 after a period time of time
	double cmpProbChange (const int stateNo1, const int stateNo2, const double time);
	//compute the first order probability of changing from stateNo1 into stateNo2 after a period time of time

	//compute the probability mat
	void cmpProbChangeDerivatives (const double time, DMat20 &probMat, DMat20 &derv1, DMat20 &derv2);

	//compute the probability mat
	void cmpProbChangeDerivatives (const int stateNo1, const double time, DVec20 &probVec, DVec20 &derv1, DVec20 &derv2);

	void cmpProbChangeDerivatives (const double brLen, int classNo, DMat20 &probMat,
        DMat20 &derv1, DMat20 &derv2);
		

	/**
		set the iteration step
	*/
	void setPamStep(int step);
	
	/**
		get the iteration step
	*/
	int getPamStep();
	
		
	//release all memmory for this class
	void release ();

	//the constructor
	virtual ~Model ();

	/************************************************************************************
	************************************************************************************
	************************************************************************************
	************************************************************************************
	************************************************************************************/

	//the type of evolutional change
	MODEL modelType_;

	PAM_TYPE baseFrqType_;
	PAM_TYPE tsTvRatioType_;
	PAM_TYPE pyPuRatioType_;
	PAM_TYPE genPamType_;


	//the transition transvertion ratio
	double tsTvRatio_;
	double pyPuRatio_;
	//double v_parm;
	
	double tsAG_;
	double tsCT_;

	double tvAC_;
	double tvAT_;
	double tvCG_;
	double tvGT_;

	MODEL 		codon_model;
    PAM_TYPE    codonFrqType;
    double      codon_tsTvRatio;
    PAM_TYPE    codon_tsTvRatioType;

	//the state frequency list

	DVec20 stateFrqArr_;

	// probability change matrix
	//DMat20 proChangeMat_;

	int nState_;
	DATA_TYPE dataType_;
	
	/**
		true of want to use POWELL method
	*/
	int pam_brent;


	BaseM     *model_;
	/*
	GTRM      gtrM_;
	HKY85M     hKY85M_;
	TN93M     tN93M_;
	AminoM     aminoM_;
	CodonModel codon_GY94;
	*/
	int nPamStep;
	bool initialized_tsTv;
	
}
; //end of class model

extern Model mymodel;

#endif
